/*
** Contains class TreeNode
**
** Part of this file was adapted from 
** https://chapel.svn.sourceforge.net/svnroot/chapel/trunk/test/studies/uts/
*/
 
module NodeModule {

  use Uts;

  /*
  ** TreeNode = node of UTS tree = task in task pool
  ** TreeNode contains methods for node expansion 
  ** and assignment of initial tasks to workers
  */
  class TreeNode {
    var height:   int;
    var hash:    [1..20] uint(8);

    // Local node expansion: children are returned via yield
    iter genChildrenHere(): TreeNode {
      var nChildren: int = 0;
      select distrib {
        when NodeDistrib.Geometric do 
          nChildren = numGeoChildren(geoDist);
        when NodeDistrib.Binomial do
          nChildren = numBinChildren();
        when NodeDistrib.Hybrid do
          nChildren = numHybridChildren();
        when NodeDistrib.Constant do
          nChildren = numConstChildren();
      }
      for i in 0 .. #nChildren {
        var child: TreeNode = new TreeNode(height + 1);
        rng_spawn(hash[1], child.hash[1], i);
        yield child;
      }
    }

    // Generates grandchildren of root and cyclically places them in
    // the various local and remote pools. This scheme is
    // simplistic - it assumes that the number of grandchildren
    // is larger than the number of workers.  
    //
    // Returns number of nodes generated thus far 
    proc generateAndDistributeInitTasks(): int {
      var whereGlobalID = 0;
      var nChildren: int = 0;
      select distrib {
        when NodeDistrib.Geometric do 
          nChildren = numGeoChildren(geoDist);
        when NodeDistrib.Binomial do
          nChildren = numBinChildren();
        when NodeDistrib.Hybrid do
          nChildren = numHybridChildren();
        when NodeDistrib.Constant do
          nChildren = numConstChildren();
      }
      for i in 0 .. #nChildren {
        var child: TreeNode = new TreeNode(1);
        rng_spawn(hash[1], child.hash[1], i);
        var nThisGrandChildren: int = 0;
        select distrib {
          when NodeDistrib.Geometric do 
            nThisGrandChildren = child.numGeoChildren(geoDist);
          when NodeDistrib.Binomial do
            nThisGrandChildren = child.numBinChildren();
          when NodeDistrib.Hybrid do
            nThisGrandChildren = child.numHybridChildren();
          when NodeDistrib.Constant do
            nThisGrandChildren = child.numConstChildren();
        }
        for ii in 0 .. #nThisGrandChildren {
          var grandChild: TreeNode = new TreeNode(2);
          rng_spawn(child.hash[1], grandChild.hash[1], ii);
          on poolComm[whereGlobalID] {
            var grandChildCopy = new TreeNode(grandChild.height, grandChild.hash);
            poolComm[whereGlobalID].push(grandChildCopy);
          }
          delete grandChild;
          whereGlobalID = (whereGlobalID + 1) % numWorkers;
        }
        delete child;
      }
      delete this;
      return nChildren + 1;
    }

    // Constant Distribution: Find the number of children
    //  Note: This is a _balanced_ tree.
    proc numConstChildren(): int {
      return if height < MAX_HEIGHT then B_0 else 0;
    }

    // Hybrid trees are geometric at the top of the tree and binomial
    // below a certain height.
    proc numHybridChildren(): int {
      if (height < shiftHeight * MAX_HEIGHT) then
        return numGeoChildren(geoDist);
      else 
        return numBinChildren();
    }

    // Binomial Distribution: Find the number of children
    //  Note: distribution is identical everywhere below root
    proc numBinChildren(): int {
      if (height == 0) then
        return B_0;
      else
        return if to_prob(rng_rand(hash[1])) < nonLeafProb then nonLeafBF else 0;
    }


    // Geometric Distribution: Find the number of children
    proc numGeoChildren(shape: GeoDistrib): int {
      var b_i: real = B_0;
  
      // use shape function to compute target b_i
      if (height > 0) {
        select shape {
      
          // Expected size is polynomial in height
          when GeoDistrib.GeoPoly do
            b_i = B_0 * (height ** (-log(B_0:real)/log(MAX_HEIGHT:real)));
        
          // Expected size is cyclic
          when GeoDistrib.GeoCyclic {
            if (height > 5 * MAX_HEIGHT) then
              b_i = 0.0;
            else
              b_i = B_0 ** sin(2.0*3.141592653589793*height/MAX_HEIGHT);
          }

          // Expected size is the same at all nodes up to max height
          when GeoDistrib.GeoFixed do
            b_i = if (height < MAX_HEIGHT) then B_0 else 0;
 
          // Expected size decreases linearly in b_i
          when GeoDistrib.GeoLinear do 
            b_i =  B_0 * (1.0 - height:real / MAX_HEIGHT:real);
        
        }
      }

      // given target b_i, find prob p so expected value of 
      // geometric distribution is b_i.
      var p: real = 1.0 / (1.0 + b_i);

      // get uniform random number on [0,1)
      var u: real = to_prob(rng_rand(hash[1]));

      // max number of children at this cumulative probability
      // (from inverse geometric cumulative density function)
      return floor(log(1 - u) / log(1 - p)):int; 
    }
  }

}




