/* 
** Worker code: main loop and task pool access functions
*/

module Worker {

  use Math; 
  use Random;
  use Pool;
  use NodeModule;

  class Worker {
  
    var globalID: int;

    var myNNodes: int = 0;

    const taskblockDomain = {0 .. #k};

    var sq: [sqDomain] TreeNode;
    var head: int = 0;
    var tail: int = 0;
    var nprivate: int = 0;
    var npublic$: sync int = 0;

    // Main loop of worker
    proc run(): uint {
      var myNNodes: uint = 0;
      while (processTask()) {
         myNNodes += 1; 
      }
      return myNNodes; 
    }

    // Function processTask includes:
    //      - pop, acquire and/or steal to get next task and
    //      - invocation of node extraction method for this task
    // Returns true if successful
    // Returns false if no work was found after ONE round of steal tries 
    // -> This is a simplistic termination scheme (see paper).
    proc processTask(): bool {
      // Acquire/steal moves task block to private pool if
      // possible. Operator || uses short cut.
      if (nprivate > 0 || tryAcquire() || tryStealLocal() || tryStealRemote()) {
        // Pool is not empty, so take a task out of the pool
        if (head == 0) {
          head = sqSize - 1;
        } else {
          head = head - 1;
        }
        nprivate -= 1;
        // Copy the parent task out of the pool.
        var task: TreeNode = sq[head];
        var nChildren = task.genChildrenHere(sq, head);
        head = (head + nChildren) % sqSize;
        nprivate += nChildren;
        while (nprivate >=  2 * k) {
          release();
        }
        return true;
      } else {
        return false;  
      }
    }

    // Tries to "move" elements from public to private pool.
    // Returns true if successful.
    proc tryAcquire(): bool {
      var np: int = npublic$;   // Begin critical section
      if (np > 0) {
        npublic$ = np - k;         // End critical section
        // Rest can be outside critical section since memory consistency 
        // model disallows statement reordering
        nprivate += k; 
        return true;
      } else {
        npublic$ = np;              // End critical section
        return false;
      }
    }

    //  Returns true if successful.
    proc tryStealLocal(): bool {   
      for victimID in localIDsCyclic() do {
        if (poolComm[victimID].npublic$.readFF() > 0) {
          var victim: Worker = poolComm[victimID];;
          var np: int = victim.npublic$;  // Begin critical section
          if (np > 0) {
            // no modulo required, since sqSize is multiple of k
            sq[taskblockDomain] = victim.sq[victim.tail..];
            victim.tail = (victim.tail + k) % sqSize;
            victim.npublic$ = np - k;    // End critical section
            nprivate = k;
            head = k;
            tail = 0;
            return true;
          } else {
             victim.npublic$ = np;       // End critical section
          }
        }
      }
      return false;
    }

   //  Returns true if successful.
    proc tryStealRemote(): bool {
      var thiefLocale: locale = here;
      var success: bool = false;   // variable resides on thiefLocale
      for victimID in remoteIDsCyclic() do {
        if (poolComm[victimID].npublic$.readFF() > 0) {
          on poolComm[victimID] {
            var victim: Worker = poolComm[victimID];
            var localCopy: [taskblockDomain] TreeNode;
            var np: int = victim.npublic$; // Begin critical section
            if (np > 0) {
              success = true;
               // no modulo required, since sqSize is multiple of k
              localCopy = victim.sq[victim.tail..];
              victim.tail = (victim.tail + k) % sqSize;
              np -= k;
            }
            victim.npublic$ = np;    // End critical section
            on thiefLocale {
              if (success) {
                sq[taskblockDomain] = localCopy;
                nprivate = k; 
                head = k; 
                tail = 0; 
              }
            }
          }
          if (success) {
              return true; 
          } 
        }  // end if
      }  // end for
      return false;
    }

    // Iterates over 0 ..numWorkersPerLoc -1, starting with startID
    // Asserts startID < numWorkersPerLoc
    iter baseCyclic(startID: int): int {
      var i: int = (startID + 1 ) % numWorkersPerLoc;
      while (i != startID) {
        yield i;
        i = (i + 1) % numWorkersPerLoc;
      }
    }

    // Iterates over globalID's of own locale
    iter localIDsCyclic(): int {
      var myRest: int = globalID % numWorkersPerLoc;
      var myStart = globalID - myRest;
      for i in baseCyclic(myRest) do {
        yield myStart + i;
      }
    }

    // Iterates over globalID's of remote locales
    iter remoteIDsCyclic(): int {
      var myRest: int = globalID % numWorkersPerLoc;
      var myStart = globalID - myRest;   // = here.id * numWorkersPerLoc
      var j: int = (myStart + numWorkersPerLoc) % numWorkers; // first ID of remote locale
      while (j != myStart) {
        yield myRest + j;
        for i in baseCyclic(myRest) do {
          yield i + j;
        }
        j = (j + numWorkersPerLoc) % numWorkers;
      }
    }

    // "Move" elements from private to public pool
    proc release() {
      // For simplicity halts if pool is full
      if (npublic$.readFF() + k  > publicSize) { 
        writeln(globalID + ": shared Pool is (almost) full");
        halt();
      }
      nprivate -= k;
      npublic$ += k;   
    }

   // Inserts one task into the local pool. Is only invoked when distributing init 
   // tasks. Thereafter tasks are expanded directly into their pool entry.
   proc push(task: TreeNode) {
      sq[head] = task;
      head += 1;
      nprivate += 1;
      if (nprivate > 2 * k) {
        release();
      }
    }

  }

}
