/* Management of tree and task pool parameters */
import x10.compiler.*;

public struct Paras {

  // uses struct, since X10 does not support enums
  public static struct NodeDistr {
    public static val Binomial = 0n;
    public static val Geometric = 1n;
    public static val Hybrid = 2n;
    public static val Constant = 3n;
  }

  public static struct GeoDistrib {
    public static val GeoFixed = 0n;
    public static val GeoLinear = 1n;
    public static val GeoPoly = 2n;
    public static val GeoCyclic = 3n;
  }

  // Parameters and names are analogous to the Chapel variant
  val seed: Int;   // Seed for the hash of the root node
  val maxChildren: Int;  // Max. Children a node may have
  val b_0: Double;  // Branching factor
  val maxHeight: Int;  //  Height limit
  val nonLeafProb: Double;    // Probability of a non-leaf (binomial)
  val nonLeafBF: Int;   // Non-leaf branching factor (binomial)
  val shiftHeight: Double; // Height at which hybrid trees go Geo=>Bin

  val distrib: Int;
  val geoDist: Int;

  val numWorkersPerPlace: Long;
  val k: Long;
  val privateSize: Long;
  val publicSize: Long;

  def this(args:Rail[String]) {
    seed = initPara(args, 'r', 16n);
    maxChildren = initPara(args, 'c', 10n);
    b_0 = initPara(args, 'b', 4.0);
    maxHeight = initPara(args, 'd', 6n);
    nonLeafProb = initPara(args, 'q', 0.10);
    nonLeafBF = initPara(args, 'm', 4n);
    shiftHeight = initPara(args, 'f', 0.5);
    distrib = initPara(args, 't', NodeDistr.Geometric);
    geoDist = initPara(args, 'a', GeoDistrib.GeoFixed);
    numWorkersPerPlace = initPara(args, 'p', 3n);
    k = initPara(args, 'k', 10);
    privateSize = initPara(args, 'x', 2 * k);
    publicSize = initPara(args, 'y', 10000 * k);
  }

  @NonEscaping def initPara(args:Rail[String], c: Char, value: Int): Int {
    var i: Long;
    for (i = 0; i < args.size && args(i).charAt(0n) != c; i += 1);
    if (i < args.size) { 
      return Int.parse(args(i).substring(1n)); 
    } else {
      return value;
    }
  }

  @NonEscaping def initPara(args:Rail[String], c: Char, value: Long): Long {
    var i: Long;
    for (i = 0; i < args.size && args(i).charAt(0n) != c; i += 1);
    if (i < args.size) { 
      return Long.parse(args(i).substring(1n)); 
    } else {
      return value;
    }
  }

  @NonEscaping def initPara(args:Rail[String], c: Char, value: Double): Double {
    var i: Long;
    for (i = 0; i < args.size && args(i).charAt(0n) != c; i += 1);
    if (i < args.size) { 
      return Double.parse(args(i).substring(1n)); 
    } else {
      return value;
    }
  }

  def showParas() {
    Console.OUT.println("UTS - Unbalanced Tree Search");
    Console.OUT.println("Tree type: " + distrib);
    Console.OUT.println("Tree shape parameters:");
    Console.OUT.println("  root branching factor b_0 = " +  b_0 + " root seed = " + seed);
    
    if (distrib == NodeDistr.Geometric || distrib == NodeDistr.Hybrid) {
      Console.OUT.println("  Geo. params: Max_Height = " + maxHeight 
                          + ", shape function = " + geoDist);
    }
    if (distrib == NodeDistr.Binomial || distrib == NodeDistr.Hybrid) {
      Console.OUT.println("  Bin. params: nonLeafProb=" + nonLeafProb 
                          + " nonLeafBF=" + nonLeafBF);
    }
    if (distrib == NodeDistr.Hybrid) {
      Console.OUT.println("  Hybrid params: shiftHeight=" + shiftHeight);
    }
    Console.OUT.println("Number of places =" + Place.numPlaces());
    Console.OUT.println();
    Console.OUT.println("Number workers per place =" + numWorkersPerPlace);
    Console.OUT.println();
  }

}