/*
** State and access functions of one split queue. The queue holds tree nodes,
** no need for generics.
*/

class SplitQueue {

  var sq: Rail[TreeNode];  
  var head: Long = 0;      // first free position
  var tail: Long = 0;      // last filled position
  var npublic: Long = 0;   
  var nprivate: Long = 0; 
  val workerID: Long;      // within place
  val tp: PlaceLocalHandle[InfosPerPlace]; 

  def this(privateSize: Long, publicSize: Long, workerID: Long, 
           tp: PlaceLocalHandle[InfosPerPlace]) {
    this.sq = new Rail[TreeNode](privateSize + publicSize);
    this.workerID = workerID;
    this.tp = tp;
  }

  // Inserts one task into the local pool. Is invoked for init and normal tasks.
  def push(var node: TreeNode) { 
    val privateSize = tp().paras.privateSize;
    if (nprivate >= privateSize) {
      release();
    }
    sq(head) = node;
    head = (head + 1) % sq.size;
    nprivate++;
  }

  // Returns next task or null.
  def pop(): TreeNode {
    // Operator || uses short cut.
    if (nprivate > 0 || tryAcquire() || tryStealLocal() || tryStealRemote()) { 
      // here is guaranteed that nprivate > 0
      nprivate--;
      head--;
      return sq(head);
    } else {
      return null;
    }
  }

  // "Move" elements from private to public pool
  def release() {
    val k = tp().paras.k;
    val publicSize = tp().paras.publicSize;
    nprivate -= k;
    atomic {
      npublic += k;
      if (npublic > publicSize) { 
        throw new Exception("release not possible: public pool full");
      }
    } 
  }

  // Tries to "move" elements from public to private pool.
  // Returns true if successful.
  def tryAcquire(): Boolean {
    val k = tp().paras.k;
    atomic {
      if (npublic >= k) {
        npublic -= k;
        nprivate += k;
        return true;
      }
    }
    return false;
  }

   //  Returns true if successful.
  def tryStealLocal(): Boolean {
    val myInfos = tp();
    val numWorkersPerPlace = myInfos.paras.numWorkersPerPlace;
    val k = myInfos.paras.k;
    var victimID: Long = (workerID + 1) % numWorkersPerPlace;
    while (victimID != workerID) {
      val victimSQ = myInfos.sqsOfOnePlace(victimID);
      atomic {
        if (victimSQ.npublic >= k) {
           victimSQ.npublic -= k;
           nprivate = k;  // may be moved outside atomic, when bool variable is used
           // may start from 0, since my pool is empty
           head = tail = 0;  
           // copy tasks
           for (var iVictimID: Long = victimSQ.tail; 
                head < k;
                head++, iVictimID++)  {
             sq(head) = victimSQ.sq(iVictimID);
           }
           victimSQ.tail = (victimSQ.tail + k) % (sq.size);
           return true;
        }
      }
      victimID = (victimID + 1) % numWorkersPerPlace;
    }
    return false;
  }

  //  Returns true if successful.
  def tryStealRemote(): Boolean {
    val numWorkersPerPlace = tp().paras.numWorkersPerPlace;
    val k = tp().paras.k;
    var victimPlace: Place = here.next();
    while (victimPlace != here) {
      val oldPlace = here;
      at (victimPlace) {
        var victimID: Long = workerID;
        // local Copy to keep atomic section short
        val stolenArray: Rail[TreeNode] = new Rail[TreeNode](k);
        var success: Boolean = false;
        do {
          val victimSQ = tp().sqsOfOnePlace(victimID);
          atomic {    // Lock is only required at victimPlace, not here
            if (victimSQ.npublic >= k) {
              success = true;
              victimSQ.npublic -= k;
              val tail = victimSQ.tail;
              for (i in 0 .. (k -1)) {
                stolenArray(i) = victimSQ.sq(i + tail);
              }
              victimSQ.tail = (tail + k) % sq.size;
            }
          }
          if (success) {
            at (oldPlace) {
              val workerSQ = tp().sqsOfOnePlace(workerID);
              for (i in 0..(k - 1)) {
                workerSQ.sq(i) = stolenArray(i);
              }
              workerSQ.tail = 0;
              workerSQ.head = k;
              workerSQ.nprivate = k;
            }
            break;
          }
          victimID = (victimID + 1) % numWorkersPerPlace;
        } while (victimID != workerID); 
      }  // end at(victimPlace) 
      if (nprivate > 0) {     // success can only be accessed at victimPlace
        return true;
      }
      victimPlace = victimPlace.next();
    }
    return false;
  }

}