/* 
** Node expansion and assignment of initial tasks to workers. 
** 
** Part of this file was adapted from 
** http:// sourceforge.net/p/x10/code/HEAD/tree/benchmarks/trunk/OBSOLETE/UTS/
*/

import x10.compiler.*;
import x10.io.Console;

class TreeNode {

  val height: Long;
  val hash: SHA1Rand;
  private static val NORMALIZER = 2147483648.0; 

  // only invoked for root of tree
  def this(height: Long, seed: Int) {
    this.height = height;
    this.hash = SHA1Rand(seed);
  }

  // invoked for non-root tree nodes
  def this (parent:TreeNode, spawnNumber:Long, height:Long) {
    this.height = height;
    this.hash = SHA1Rand (parent.getHash(), spawnNumber as Int);
  }

  public operator this (): Int = this.hash();

  def getHash(): SHA1Rand = this.hash;

   // Generates grandchildren of root and cyclically places them in
   // the various local and remote pools. This scheme is
   // simplistic - it assumes that the number of grandchildren
   // is larger than the number of workers.  
   //
   // Returns number of nodes generated thus far 
  def initPool(tp: PlaceLocalHandle[InfosPerPlace]): Int {
    var place: Place = here;
    val paras = tp().paras;
    val nChildren: Long;
    switch (paras.distrib) {
      case Paras.NodeDistr.Binomial: nChildren = paras.b_0 as Long; break;  // special case for root
      case Paras.NodeDistr.Geometric: nChildren = numGeoChildren(paras); break;
      default: throw new Exception("NodeDistr not implemented");
    }
    val childArray: Rail[TreeNode] = new Rail[TreeNode](nChildren,  (i: long) => new TreeNode(this, i, 1));
    var nGrandchildrenArray: Rail[Long] = new Rail[Long](nChildren);
    for (i in 0 .. (nChildren - 1)) {
      switch (paras.distrib) {
        case Paras.NodeDistr.Binomial: 
          nGrandchildrenArray(i) = childArray(i).numBinChildren(paras); 
          break;
        case Paras.NodeDistr.Geometric: 
          nGrandchildrenArray(i) = childArray(i).numGeoChildren(paras); 
          break;
        default: throw new Exception("NodeDistr not implemented");
      }
    }
    var nTotalGrand: Long = 0;
    for (i in 0 .. (nChildren - 1)) {
      nTotalGrand += nGrandchildrenArray(i);
    }
    val tasksPerPlace: Long = nTotalGrand / Place.MAX_PLACES;
    val surplus: Long = nTotalGrand % Place.MAX_PLACES;
    var actChild: Long = 0;
    var actGrandchild: Long = 0;
    var tasksAssigned: Long = 0;   // how many have already been assigned to actual place
    while (actChild < nChildren) {
      val eps = place.id() < surplus ? 1 : 0;
      var howManyNow: Long = Math.min(tasksPerPlace + eps - tasksAssigned, 
                                                                          nGrandchildrenArray(actChild) - actGrandchild);
      val valActChild = actChild;
      val valActGrandchild = actGrandchild;
      val valNumWorkersPerPlace = paras.numWorkersPerPlace;
      val startLocalWorkerID = tasksAssigned % valNumWorkersPerPlace;
      val endGrandchild = actGrandchild + howManyNow;  // next unassigned
      at (place) {
        var workerID: Long = startLocalWorkerID;
        for (j in valActGrandchild .. (endGrandchild - 1)) {
          val grandChild: TreeNode = new TreeNode(childArray(valActChild), j, 2);
          tp().sqsOfOnePlace(workerID).push(grandChild);
          workerID = (workerID + 1) % valNumWorkersPerPlace; 
        }
      }
      if (howManyNow == tasksPerPlace + eps - tasksAssigned) { 
        place = place.next(); 
        actGrandchild = endGrandchild;
        tasksAssigned = 0;
        howManyNow = 0;
      } 
      if (howManyNow == nGrandchildrenArray(actChild) - actGrandchild) {
        actChild++;
        actGrandchild = 0;
        tasksAssigned += howManyNow;
      }
    }
    return (nChildren + 1) as Int;
  }

  // Local node expansion, invokes push to insert children into local pool
  def genChildrenHere(workerID: Long, tp: PlaceLocalHandle[InfosPerPlace]) {
    val paras = tp().paras;
    val workerSQ = tp().sqsOfOnePlace(workerID);
    val nChildren: Long;
    switch (paras.distrib) {
      case Paras.NodeDistr.Binomial: nChildren = numBinChildren(paras); break;
      case Paras.NodeDistr.Geometric: nChildren = numGeoChildren(paras); break;
      default: throw new Exception("NodeDistr not implemented");
    }
    for (i in 0 .. (nChildren - 1)) {
       workerSQ.push(new TreeNode(this, i, height + 1));
    }
  }

  private def numGeoChildren (paras: Paras): Int {
    var curNodeBranchingFactor:double;
    if (height == 0) {
      curNodeBranchingFactor = paras.b_0;
    } else {
      switch (paras.geoDist) {
      case Paras.GeoDistrib.GeoPoly: 
         val tmpLogOne = -1.0 * Math.log (paras.b_0 as double);
         val tmpLogTwo = Math.log (paras.maxHeight as double);
         curNodeBranchingFactor = paras.b_0 
                                  *  Math.pow(height as double,tmpLogOne/tmpLogTwo);
         break;
       case Paras.GeoDistrib.GeoCyclic:
         if (height > 5 * paras.maxHeight) {
           curNodeBranchingFactor = 0.0;
         } else {
           val exponent = Math.sin(2.0 * Math.PI 
                          * (height as double) / (paras.maxHeight as double));
           curNodeBranchingFactor = Math.pow (paras.b_0, exponent);
         };
         break;
       case Paras.GeoDistrib.GeoFixed:
         curNodeBranchingFactor = (height < paras.maxHeight) ? paras.b_0: 0.0;
         break;
       case Paras.GeoDistrib.GeoLinear:
         curNodeBranchingFactor = paras.b_0 
                                  * (1.0 - (height as double) / (paras.maxHeight as double));
         break;
       default:
         curNodeBranchingFactor = 0;
         Console.OUT.println ("Unknown shape function for geometric UTS");
       }
     }
    val probForCurNodeBranchingFactor = 1.0 / (1.0 + curNodeBranchingFactor);
    val randomNumber = this() as double;
    val normalizedRandomNumber = randomNumber / 2147483648.0;
    return Math.floor ((Math.log (1 - normalizedRandomNumber))
           / (Math.log (1 - probForCurNodeBranchingFactor))) as int;
  }
    
  private def numBinChildren(paras: Paras): Int {
    val randomNumber = this() as double;
    val normalizedRandomNumber = randomNumber / 2147483648.0;
    return (normalizedRandomNumber < paras.nonLeafProb) ? paras.nonLeafBF : 0n;
  }      
}

